/*
	dash rings, written and sprited by Lighto
*/

local ver, rev = 1, 16

freeslot("SPR_LBHD","MT_LTB_HORIDASHRNG",
		 "SPR_LBVD","MT_LTB_VERTDASHRNG",
		 "SPR_LBDD","MT_LTB_DIAGDASHRNG",
		 "S_LTB_HORIDASHRINGCOMM",
		 "S_LTB_VERTDASHRINGCOMM",
		 "S_LTB_DIAGDASHRINGCOMM")

mobjinfo[MT_LTB_HORIDASHRNG] = {
	--$Name Horizontal Dash Ring
    --$Sprite LBHDA3A7
    --$Category Lighto Tool Box
	--$Flags1Text Centralize
	--$Flags4Text keep speed
	--$Flags8Text add speed	
	--$ParameterText speed
    doomednum = 2450,
    spawnstate = S_LTB_HORIDASHRINGCOMM,
    spawnhealth = 1000,
    radius = 80*FRACUNIT,
    height = 140*FRACUNIT,
    mass = 0,
    damage = 1,
	speed = 0,
    flags = MF_SOLID|MF_NOGRAVITY|MF_DONTENCOREMAP
}
mobjinfo[MT_LTB_VERTDASHRNG] = {
	--$Name Vertical Dash Ring
    --$Sprite LBVDA0
    --$Category Lighto Tool Box
	--$Flags1Text Centralize
	--$Flags4Text keep speed
	--$Flags8Text add speed	
	--$ParameterText speed
    doomednum = 2451,
    spawnstate = S_LTB_VERTDASHRINGCOMM,
    spawnhealth = 1000,
    radius = 80*FRACUNIT,
    height = 120*FRACUNIT,
    mass = 1,
    damage = 0,
	speed = 0,
    flags = MF_SOLID|MF_NOGRAVITY|MF_DONTENCOREMAP
}
mobjinfo[MT_LTB_DIAGDASHRNG] = {
	--$Name Diagonal Dash Ring
    --$Sprite LBDDA3A7
    --$Category Lighto Tool Box
	--$Flags1Text Centralize
	--$Flags4Text keep speed
	--$Flags8Text add speed	
	--$ParameterText speed
    doomednum = 2452,
    spawnstate = S_LTB_DIAGDASHRINGCOMM,
    raisestate = S_NULL,
    spawnhealth = 1000,
    radius = 80*FRACUNIT,
    height = 120*FRACUNIT,
    mass = 1,
    damage = 1,
	speed = 0,
    flags = MF_SOLID|MF_NOGRAVITY|MF_DONTENCOREMAP
}

states[S_LTB_HORIDASHRINGCOMM] = {SPR_LBHD,A,-1, nil, 0,0 ,S_NULL }
states[S_LTB_VERTDASHRINGCOMM] = {SPR_LBVD,A,-1, nil, 0,0 ,S_NULL }
states[S_LTB_DIAGDASHRINGCOMM] = {SPR_LBDD,A,-1, nil, 0,0 ,S_NULL }


local GREY, YELLOW, RED, BLUE = 2, 47, 22, 85 --colors
local EXTRA, FLIP, SPECIAL, AMBUSH = 1, 2, 4, 8 --yes
local COLOR = {GREY, YELLOW, RED, BLUE}
local COLORBLINK = {SKINCOLOR_WSUPER2, SKINCOLOR_SUPER1, SKINCOLOR_LEMONADE, SKINCOLOR_SLATE} --todo, correct colors
local dashringlist = {MT_LTB_HORIDASHRNG, MT_LTB_VERTDASHRNG, MT_LTB_DIAGDASHRNG}

if LTB_dashrings == nil or not LTB_dashrings then
	rawset(_G, "LTB_dashrings", {}) --ayo
end

--basic dash ring setup
--each dash ring is given it's power
--and change its color based on that
local function localdashringthinker(dr)
	if not(dr and dr.valid) then return end --paranoia????
	if not dr.LTB_dashringinit and dr.spawnpoint then
		dr.LTB_dashringinit = true
		dr.LTB_dashringtier = 1
		--had to do that way >:\
		if dr.spawnpoint.extrainfo > 12 then
			dr.LTB_dashringtier = 4
		elseif dr.spawnpoint.extrainfo > 8 then
			dr.LTB_dashringtier = 3
		elseif dr.spawnpoint.extrainfo > 4 then
			dr.LTB_dashringtier = 2
		end
		dr.color = COLOR[dr.LTB_dashringtier]
		dr.LTB_DR_mass = 0
		dr.LTB_DR_damage = 0
		local basepower = 22 * mapobjectscale
		if dr.info.mass then 
			dr.LTB_DR_mass = basepower + ((8*mapobjectscale) * dr.spawnpoint.extrainfo  ) 			
		end
		if dr.info.damage then 
			dr.LTB_DR_damage = basepower + ((8*mapobjectscale) * dr.spawnpoint.extrainfo  ) 
		end
		
		dr.LTB_istouchedcooldown = 0
		dr.LTB_isverticaldashring = dr.info.mass and not dr.info.damage
		dr.LTB_ishorizontaldashring = dr.info.damage and not dr.info.mass
		dr.LTB_isdiagonaldashring = dr.info.mass and dr.info.damage
		dr.LTB_isadashring = true
	end	
	
	//make them blink
	if dr.LTB_istouchedcooldown > 0 then
		dr.color = COLORBLINK[dr.LTB_dashringtier]
		dr.colorized = true		
	else
		dr.color = COLOR[dr.LTB_dashringtier]
		dr.colorized = false
	end
	
	
	dr.LTB_istouchedcooldown = max(0, ($ or 0) - 2)
end


--cannot touch the same dashring twitch without touching another one or die/bump/spinout/get squished
--collision with the dashrings
--only meant to interact with players as a mean of traveling through the air
--which the base game lacks of (speed pads can work but are janky and unnatural)
--theres three ways to change each one
--EXTRA will make players center to the dashring
--AMBUSH will keep the current momentum, only changing the moving direction
--SPECIAL will add the power to the current speed
--those allow to drift store and it wanna give some small vertical speed to keep players in the air for more time
--in addition to this, since the game has a very brute speed cap on ground
--abruptly slows the player down, when landing after hitting a dashring, it will remove some friction ,amking the player slide, making it smoother
local function localdashringcollision(dashring, thing  )
	if thing.z > dashring.z + dashring.height
	or thing.z + thing.height < dashring.z then return false end
	if thing.LTB_toucheddashring == dashring then 
		thing.LTB_dashringcooldown = ($ or 0) + 2
		return false 
	end
	if thing.eflags & MFE_SPRUNG then return false end
	//if not thing.player then return false end
	
	local opt = dashring.spawnpoint.options
	local horispeed, vertspeed = dashring.LTB_DR_damage, dashring.LTB_DR_mass * P_MobjFlip(dashring)
	
	--keep momentum
	if opt & SPECIAL then
		if not dashring.LTB_isverticaldashring then
			horispeed = max(FixedHypot(thing.momx, thing.momy),26*mapobjectscale)	
		else
			vertspeed = max(thing.momz* P_MobjFlip(dashring),26*mapobjectscale* P_MobjFlip(dashring))
		end
	end
	
	--add momentum to the dash ring power
	if opt & AMBUSH then
		if not dashring.LTB_isverticaldashring then
			horispeed = $ + max(FixedHypot(thing.momx, thing.momy),26*mapobjectscale)	
		else
			vertspeed = $ + max(thing.momz* P_MobjFlip(dashring),26*mapobjectscale* P_MobjFlip(dashring))
		end
	end
	
	thing.eflags = $ | MFE_SPRUNG
	thing.momx, thing.momy, thing.momz = 0,0,0
	
	--look in the same direction as the dashring
	
	
	--move to the dashring
	local mx, my, mz = thing.x , thing.y, thing.z 
	if opt & EXTRA then --centralize
		if not dashring.LTB_isverticaldashring then
			mx = dashring.x 
			my = dashring.y 
			mz = dashring.z - ((16*mapobjectscale) * P_MobjFlip(thing))  - min(0,(dashring.height - thing.height )* P_MobjFlip(thing))

			if dashring.LTB_isdiagonaldashring then
				mz = dashring.z - ((16*mapobjectscale) * P_MobjFlip(thing)) - min(0,(dashring.height - thing.height )* P_MobjFlip(thing))
			else
				vertspeed = $ + (-P_GetMobjGravity(thing) *8)
			end
			
			thing.LTB_vertdashrngdashcnt = 0
		
		else
			mx = thing.x
			my = thing.y
			mz = thing.z - ((((dashring.height/2) + (6*mapobjectscale)))* P_MobjFlip(thing)) - min(0,(dashring.height - thing.height )* P_MobjFlip(thing))
			
			thing.LTB_vertdashrngdashcnt = ($ or 0) + 1
			
			--if you bounce again on the same dahsring, push the player out, maybe ot stuck by an explosion???
			if thing.LTB_vertdashrngdashcnt
			and thing.LTB_vertdashrngdashcnt > 1 then
				thing.LTB_vertdashrngdashcnt = 0
				horispeed = 16*mapobjectscale
			end	
		end
		
		
	end
	mz = $ -P_GetMobjGravity(thing)*TICRATE --a little above the ground
	thing.flags = $ | (MF_NOCLIP|MF_NOCLIPHEIGHT) --SAY BYE TO THE WEIRD SLOPE PHYSICS	
	--de-solidify
	--move to the spring (or to front)
	--give the thrust
	dashring.flags = $ & ~(MF_SOLID|MF_SPECIAL)
	thing.z = mz
	P_TryMove(thing, mx, my, true)
	P_Thrust(thing, dashring.angle, horispeed)
	thing.momz = vertspeed 
	thing.flags = $ & ~(MF_NOCLIP|MF_NOCLIPHEIGHT) 
	dashring.flags = dashring.info.flags
	dashring.LTB_istouchedcooldown = 42
	thing.LTB_toucheddashring = dashring
	thing.LTB_dashringcooldown = 3
	S_StartSound(thing, sfx_nbmper, nil)
	if thing.player then 
		P_ResetPlayer(thing.player) 	
		if not dashring.LTB_isverticaldashring then
			--wanna remove a bit of friction to make the plaer slide for 2 secconds, for a smooth landing
			thing.LTB_dashringairspeed = horispeed
		end
	end
	return true
end


local function localplayerthink(mo)
	if not(mo and mo.valid) then return end
	local p = mo.player
	--get around the very rigid air speed cap
	if not P_IsObjectOnGround(mo) 
	and mo.LTB_toucheddashring 
	and mo.LTB_dashringairspeed
	and p.speed >= 70*mapobjectscale
	and not (FZeroKart and FZeroKart.Active) then
		p.speed = min(mo.LTB_dashringairspeed, 200*mapobjectscale)		
	end
	
	--if you bump a obstacle, or someone, or get damaged , die or for some reason, go too slow , stop the momentun AND wont change friction upon landing
	if mo.eflags & MFE_JUSTBOUNCEDWALL
	or p.kartstuff[k_spinouttimer]
	or p.kartstuff[k_squishedtimer]
	or p.kartstuff[k_justbumped]
	or p.kartstuff[k_dashpadcooldown]
	or p.speed < 4*mapobjectscale
	or p.health < 0
	or p.deadtimer then
		mo.LTB_dashringairspeed = nil
		mo.LTB_dashringcancelled = true
	end
	
	--cancel dashring speed cap prevention AND make player slide a bit to smoothen the ground speed cap upon landing
	if P_IsObjectOnGround(mo) 
	and mo.LTB_toucheddashring then
		mo.LTB_toucheddashring = nil
		mo.LTB_dashringairspeed = nil
		mo.LTB_vertdashrngdashcnt = 0
		if not mo.LTB_dashringcancelled then
			mo.LTB_dashringlandslide = FRACUNIT*3/4
		end
		mo.LTB_dashringcancelled = nil
	end
	
	--adds extra friction for 2 seconds
	if mo.LTB_dashringlandslide then
		if P_IsObjectOnGround(mo) then
			mo.LTB_dashringfricdiff = FRACUNIT - mo.friction
			mo.friction = $ + FixedMul(mo.LTB_dashringfricdiff, mo.LTB_dashringlandslide )
		end
		mo.LTB_dashringlandslide = max(0, $-FRACUNIT/2/TICRATE)
	end
	
	if mo.LTB_dashringcooldown 
	and mo.LTB_dashringcooldown < 1 then
		mo.LTB_toucheddashring = nil
	end
	
	mo.LTB_dashringcooldown = max(0, ($ or 0) - 1)
	
	
end

local function localplayerreset()
	for p in players.iterate do
		if not (p and p.valid and p.mo and p.mo.valid) then continue end
		p.mo.LTB_toucheddashring = nil
		p.mo.LTB_dashringairspeed = nil
		p.mo.LTB_dashringlandslide = nil
		p.mo.LTB_dashringcancelled = nil
		p.mo.LTB_vertdashrngdashcnt = nil
	end
end

if LTB_dashrings.ver == nil
or (LTB_dashrings.ver < ver
or (LTB_dashrings.ver == ver and (LTB_dashrings.rev == nil or LTB_dashrings.rev < rev))) then

if LTB_dashrings.ver
and (LTB_dashrings.ver < ver
or (LTB_dashrings.ver == ver and (LTB_dashrings.rev == nil or LTB_dashrings.rev < rev))) then
	print("Newer version detected, updtaing it - Dash rings by Lighto")
	LTB_dashrings.updatedfrom = {LTB_dashrings.ver, LTB_dashrings.rev}
	LTB_dashrings.updatedto = {ver, rev}
	LTB_dashrings.updated = true
	
	LTB_dashrings.ver = ver
	LTB_dashrings.rev = rev
	LTB_dashrings.funcs.dashringthinker = localdashringthinker
	LTB_dashrings.funcs.dashringcollision = localdashringcollision
	LTB_dashrings.funcs.playerthink = localplayerthink
	LTB_dashrings.funcs.playerreset = localplayerreset
else 
	print("Dash rings lua, witten by Lighto") 
		

	LTB_dashrings.ver = ver
	LTB_dashrings.rev = rev
	LTB_dashrings.init = true
	LTB_dashrings.funcs = {}
	LTB_dashrings.funcs.dashringthinker = localdashringthinker
	LTB_dashrings.funcs.dashringcollision = localdashringcollision
	LTB_dashrings.funcs.playerthink = localplayerthink
	LTB_dashrings.funcs.playerreset = localplayerreset


	for k, v in pairs(dashringlist) do
		addHook("MobjThinker", LTB_dashrings.funcs.dashringthinker, v)
		addHook("MobjCollide", LTB_dashrings.funcs.dashringcollision , v)		
	end
	addHook("MobjThinker", LTB_dashrings.funcs.playerthink, MT_PLAYER)
	addHook("MapChange",LTB_dashrings.funcs.playerreset)
end

end
